#include <stdio.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <linux/cdrom.h>
#define BUF_SIZE 75*2352
typedef struct wav_header {
    char riff[4];
    long filesize;
    char rifftype[4];
    char chunk_id1[4];
    long chunksize1;
    short wFormatTag;
    short nChannels;
    long nSamplesPerSec;
    long nAvgBytesPerSec;
    short nBlockAlign;
    short wBitsPerSample;
    char chunk_id2[4];
    long chunksize2;
} wav_header;
int msf_to_frames(struct cdrom_msf0 * msf) {
  return (msf->minute * 60 + msf->second) * 75 + msf->frame;
}
void frames_to_msf(int frames, struct cdrom_msf0 * msf) {
  msf->frame = frames % 75;
  msf->second = (frames / 75) % 60;
  msf->minute = (frames / 75) / 60;
}
int main (int argc, char ** argv) {
  int cdd, fd, track, start, stop, disc_stat;
  char buf[BUF_SIZE];
  struct cdrom_tochdr toc;
  struct cdrom_tocentry entry;	
  struct cdrom_read_audio rdaudio;
  struct wav_header hdr;
  if (argc != 3) {
    printf("Использование: %s <трек> <аудиофайл>\n", argv[0]);
    return 1;
  }
  cdd = open("/dev/cdrom", O_RDONLY);
  track = atoi(argv[1]);
  disc_stat = ioctl(cdd, CDROM_DRIVE_STATUS, CDSL_CURRENT);
  if ((disc_stat != CDS_DISC_OK) && (disc_stat != CDS_NO_INFO) && (disc_stat != -1))
  {
    close(cdd);
    printf("Устройство не готово\n");
    return 1;
  }
  ioctl(cdd, CDROMREADTOCHDR, &toc);
  if ((track < toc.cdth_trk0) || (track > toc.cdth_trk1)) {
    close(cdd);
    printf("Неверный номер трека\n");
    return 1;
  }
  entry.cdte_format = CDROM_MSF;
  entry.cdte_track = track;
  ioctl(cdd, CDROMREADTOCENTRY, &entry);
  if ((entry.cdte_ctrl & CDROM_DATA_TRACK) != 0) {
    close(cdd);
    printf("Выбран не аудио-трек\n");
    return 1;
  }
  start = msf_to_frames(&entry.cdte_addr.msf);
  entry.cdte_track = (track < toc.cdth_trk1) ? track + 1 : CDROM_LEADOUT;
  ioctl(cdd, CDROMREADTOCENTRY, &entry);
  stop = msf_to_frames(&entry.cdte_addr.msf);
  fd = open(argv[2], O_WRONLY|O_CREAT, 0777);
  memcpy(hdr.riff , (const void *) "RIFF", 4);
  memcpy(hdr.rifftype, (const void *) "WAVE", 4);
  memcpy(hdr.chunk_id1, (const void *) "fmt ", 4);
  hdr.chunksize1 = 16;
  hdr.wFormatTag = 1;  // WAVE_FORMAT_PCM;
  memcpy(hdr.chunk_id2, (const void *) "data", 4);
  hdr.nChannels = 2;
  hdr.nSamplesPerSec = 44100;
  hdr.nBlockAlign = 4;
  hdr.nAvgBytesPerSec = 44100 * hdr.nBlockAlign;
  hdr.wBitsPerSample = 16;
  hdr.chunksize2 = (stop-start)*2352;
  hdr.filesize = hdr.chunksize2 + 44;
  write(fd, &hdr, sizeof(hdr));
  rdaudio.addr_format = CDROM_MSF;
  rdaudio.buf = buf;
  while (start < stop) {
    frames_to_msf(start, &rdaudio.addr.msf);
    start += (rdaudio.nframes = (stop - start) > 75 ? 75 : stop - start);
    ioctl(cdd, CDROMREADAUDIO, &rdaudio);
    write(fd, buf, rdaudio.nframes*2352);
  }
  close(cdd);
  close(fd);
}
